static char rcsid[] = "$Author: djh $ $Date: 91/03/14 13:21:21 $";
static char rcsident[] = "$Header: lwsrv.c,v 2.3 91/03/14 13:21:21 djh Exp $";
static char revision[] = "$Revision: 2.3 $";

/*
 * lwsrv - UNIX AppleTalk spooling program: act as a laserwriter
 *  driver: handles setup and farms out incoming jobs
 *
 * AppleTalk package for UNIX (4.2 BSD).
 *
 * Copyright (c) 1986, 1987, 1988 by The Trustees of Columbia University in the
 * City of New York.
 *
 * Edit History:
 *
 *  Feb  15, 1987	Schilit		Created, based on lsrv
 *  Mar  17, 1987	Schilit		Fixed nonprintables, added -r
 *  May  15, 1987	CCKim		Add support for LaserPrep 4.0
 *					Make multifork by default (turn
 *					off by defining SINGLEFORK)
 *  Feb	 15, 1991	djh		Various cleanups & patches
 *  Feb  20, 1991	rapatel		improve wait code, add lpr logging
 *
 */

/* PATCH: Rutgers1/lwsrv.c.diffs, djh@munnari.OZ.AU, 16/11/90 */
/* PATCH: XENIX/file.3, *UNTESTED* djh@munnari.OZ.AU, 20/11/90 */

char copyright[] = "Copyright (c) 1986, 1987, 1988 by The Trustees of Columbia University in the City of New York";

#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <signal.h>
#include <sys/param.h>
#ifndef _TYPES
# include <sys/types.h>			/* assume included by param.h */
#endif  _TYPES
#include <sys/file.h>
#include <sys/ioctl.h>
#ifdef xenix5
#include <sys/ndir.h>
#else xenix5
#include <sys/dir.h>
#endif xenix5
#include <sys/stat.h>

#include <netat/appletalk.h>		/* include appletalk definitions */
#include <netat/compat.h>
#ifdef USESTRINGDOTH
# include <string.h>
#else  USESTRINGDOTH
# include <strings.h>
#endif USESTRINGDOTH
#ifndef NOWAIT3
# include <sys/wait.h>
#endif  NOWAIT3
#ifdef NEEDFCNTLDOTH
# include <fcntl.h>
#endif NEEDFCNTLDOTH
#include "papstream.h"
#ifdef LWSRV_AUFS_SECURITY
#include <pwd.h>			/* budd */
#endif LWSRV_AUFS_SECURITY
#ifdef AUTHENTICATE
#include <sys/time.h>
#include <netinet/in.h>
#include "../../lib/cap/abpap.h"	/* urk, puke, etc */
#endif AUTHENTICATE

private char *tracefile = NULL;
private char *fontfile = NULL;
private char *logfile = NULL;
private char *prtname = NULL;
private char *unixpname = NULL;
private char *prttype = "LaserWriter";
private char *dictdir = ".";		/* assume local dir */
private int rflag = FALSE;		/* remove print file */
private int hflag = TRUE;		/* default to print banner */
private int singlefork = FALSE;
export PAPStatusRec statbuff;		/* status buffer */
#ifdef PAGECOUNT
export int pagecount;
export int pcopies;
#endif PAGECOUNT

private char username[80],jobname[80];
private char buf[100];
private int srefnum;

#ifndef TEMPFILE
# define TEMPFILE "/tmp/lwsrvXXXXXX"	/* temporary file holds job */
#endif TEMPFILE

#ifdef LWSRV_AUFS_SECURITY
private int requid, reqgid;
private char *aufsdb = NULL;		/* budd */
#endif LWSRV_AUFS_SECURITY
#ifdef LWSRV_LPR_LOG
private char *requname = NULL;
#endif LWSRV_LPR_LOG
#define RFLOWQ atpMaxNum
#define BUFMAX (PAPSegSize*RFLOWQ)+10
#ifdef LWSRV_AUFS_SECURITY
int tempbinlen;
char *bin, *tempbin;
#endif LWSRV_AUFS_SECURITY
#ifdef NeXT
char *nextdpi = NULL;			/* NeXT printer resolution */
#endif NeXT

private void
usage(s,err) 
char *s,*err;
{
  if (err != NULL)
    fprintf(stderr,"%s: %s\n",s,err);
  fprintf(stderr,"usage: %s -n <PrinterName> -p <unix printer name>\n",s);
  fprintf(stderr,"usage:\t\t-a <DictionaryDirectory> -f <FontFile>\n");
  fprintf(stderr,"usage:\t\t[-l <LogFile>] [-t <TraceFile>] [-r] [-h]\n");
#ifndef LWSRV_AUFS_SECURITY
  fprintf(stderr,"usage:\t\t[-T crtolf] [-T quote8bit]\n\n");
#else LWSRV_AUFS_SECURITY
  fprintf(stderr,"usage:\t\t[-T crtolf] [-T quote8bit] [-X userlogindb]\n\n");
#endif LWSRV_AUFS_SECURITY
  fprintf(stderr,"\t-p*<unix printer name> is the unix printer to print to\n");
  fprintf(stderr,"\t-n*<PrinterName> specifies the name lwsrv registers.\n");
  fprintf(stderr,"\t-a*<DictionaryDirectory> is the ProcSet directory.\n");
  fprintf(stderr,"\t-f*<FontFile> contains an font coordination list.\n");
  fprintf(stderr,"\t-t <Tracefile> stores session and appledict in\n");
  fprintf(stderr,"\t   <Tracefile> without printing.\n");
  fprintf(stderr,"\t-l <LogFile> specifies a file to log the lwsrv session\n");
  fprintf(stderr,"\t-e Allow an eexec to occur in a procset\n");
  fprintf(stderr,"     warning: this may cause problems, use carefully\n");
  fprintf(stderr,"\t-r Will retain temp print file for inspection\n");
  fprintf(stderr,"\t-h means to print without a banner page\n");
  if (is_simple_dsc()) 
    fprintf(stderr,"\t-A [on*|off] means to turn on or off Adobe document\n");
  else
    fprintf(stderr,"\t-A [on|off*] means to turn on or off Adobe document\n");
  fprintf(stderr,"\tstructuring revision 2 compatibility\n");
  fprintf(stderr,"\t(this can cause problems, *'ed item is default)\n");
#ifdef NeXT
  fprintf(stderr,"\t-R <dpi> Specify resolution for NeXT printer\n");
#endif NeXT
  fprintf(stderr,"\t-S single lwsrv fork (default is multiforking)\n");
  fprintf(stderr,"\t-T Transcript compatibilty options\n");
  fprintf(stderr,"\t-T crtolf: translate cr to lf for Transcript filters\n");
  fprintf(stderr,"\t-T quote8bit: quote 8 bit chars for Transcript\n");
  fprintf(stderr,"\t-T makenondscconformant: make non DSC conformant: use\n");
  fprintf(stderr,"\t   if psrv only works with DSC 1.0 conventions\n");
#ifdef LWSRV_AUFS_SECURITY
 fprintf(stderr,"\t-X <database directory> use aufs for validation\n");/*budd*/
#endif LWSRV_AUFS_SECURITY
  fprintf(stderr,"\nexample: %s -n Laser -p ps -a/usr/lib/ADicts\n",s);
  fprintf(stderr,"\t\t-f /usr/lib/LW+Fonts\n");
  fprintf(stderr,"\t(note the starred items above are required)\n");
  exit(0);
}

private void
doargs(argc,argv)
int argc;
char **argv;
{
  int c;
  extern char *optarg;
  extern int optind;
  
#ifndef LWSRV_AUFS_SECURITY
#ifdef NeXT
  while ((c = getopt(argc,argv,"a:f:l:p:P:t:d:n:rehT:A:S:R")) != EOF) {
#else  NeXT
  while ((c = getopt(argc,argv,"a:f:l:p:P:t:d:n:rehT:A:S")) != EOF) {
#endif NeXT
#else LWSRV_AUFS_SECURITY
#ifdef NeXT
  while ((c = getopt(argc,argv,"a:f:l:p:P:t:d:n:rehT:A:SX:R")) != EOF) {
#else  NeXT
  while ((c = getopt(argc,argv,"a:f:l:p:P:t:d:n:rehT:A:SX:")) != EOF) {
#endif NeXT
#endif LWSRV_AUFS_SECURITY
    switch (c) {
    case 'a':
      if (index(optarg, '/') == NULL) {
	dictdir = (char *)malloc(strlen(optarg)+4);
	strcpy(dictdir, "./");
	strcat(dictdir, optarg);
      } else 
	dictdir = optarg;	/* remember dictionary directory */
      break;
    case 'd':
      dbugarg(optarg);
      break;
    case 'f':
      fontfile = optarg;		/* -f fontfile */
      break;
    case 'l':				/* -l logfile */
      logfile = optarg;
      break;
    case 'n':				/* lwsrv printer name */
      prtname = optarg;
      break;
    case 'p':				/* -p unix printer name */
    case 'P':
      unixpname = optarg;
      break;
    case 'h':
      hflag = FALSE;			/* do not print banner */
      break;
    case 'r':				/* do not remove file */
      rflag = TRUE;
      break;
    case 't':				/* -t tracefile */
      tracefile = optarg;
      break;
    case 'e':
      setflag_encrypted_instream(TRUE);	/* maybe "eexec" in code */
      break;
    case 'T':
      if (simple_TranscriptOption(optarg) < 0)
	usage(argv[0], NULL);
      break;
    case 'A':
      if (simple_dsc_option(optarg) < 0)
	usage(argv[0], NULL);
      break;
    case 'S':
      singlefork = TRUE;
      fprintf(stderr, "lwsrv: single fork\n");
      break;
#ifdef LWSRV_AUFS_SECURITY
    case 'X':				/* budd... */
      aufsdb = optarg;
      fprintf(stderr, "lwsrv: aufs db directory in %s\n", aufsdb );
      break;				/* ...budd */
#endif LWSRV_AUFS_SECURITY
#ifdef NeXT
    case 'R':
      nextdpi = optarg;
      break;
#endif NeXT
    case '?':				/* illegal character */
      usage(argv[0],NULL);		/* usage and exit */
      break;
    }
  }

  if (unixpname == NULL)
    usage(argv[0],"No Unix Printer Name");
  if (prtname == NULL)
    usage(argv[0],"No AppleTalk Printer Name");
  if (fontfile == NULL)
    usage(argv[0],"No FontFile specified");    
}

private void
childdone()
{
#ifndef NOWAIT3
  union wait status;
  while (wait3(&status, WNOHANG, 0) > 0) ;
  signal(SIGCHLD, childdone);
#else   NOWAIT3
  int status;
  (void)wait(&status);
  signal(SIGCHLD, childdone);
#endif NOWAIT3
}

/*
 * delete the NBP name
 */

private int
cleanup()
{
	SLClose(srefnum);		/* close server for child */
	PAPRemName(srefnum, buf);
	exit(0);
}

main(argc,argv)
int argc;
char **argv;
{
  int err,i,cno;
  int rcomp;
  int cleanup();
  void childdone();

  doargs(argc,argv);			/* handle args */
  
#ifdef LWSRV_AUFS_SECURITY
  requid = getuid();			/* budd */
  reqgid = getgid();			/* budd */
#endif LWSRV_AUFS_SECURITY

#ifdef AUTHENTICATE
  initauthenticate(*argv, prtname);
#endif AUTHENTICATE

  signal(SIGHUP, cleanup);
  signal(SIGINT, cleanup);
  signal(SIGQUIT, cleanup);
  signal(SIGTERM, cleanup);

  if (!dbug.db_flgs) {			/* disassociate */
    if (fork())
      exit(0);				/* kill off parent */
    for (i=0; i < 20; i++)
      close(i);				/* close all files */
    (void) open("/dev/null",0);
#ifndef NODUP2
    (void) dup2(0,1);
#else   NODUP2
    (void)dup(0);			/* for slot 1 */
#endif  NODUP2
    if (logfile == NULL) {
#ifndef NODUP2
      (void) dup2(0,2);
#else   NODUP2
      (void) dup(0);			/* for slot 2 */
#endif  NODUP2
    } else {
      if ((i = open(logfile,O_WRONLY|O_APPEND)) < 0)
	i = creat(logfile,0666);
      if (i != 2) {
#ifndef NODUP2
	(void) dup2(i,2);
#else   NODUP2
	close(2);			/* try again */
	(void) dup(0);			/* for slot 2 */
#endif  NODUP2
	(void) close(i);
      }
#ifdef LWSRV_LPR_LOG	/* Set up logfile to be also stdout/stderr for lpr */
#ifdef NODUP2
      close(1);
      dup(2);
#else  NODUP2
      dup2 (2, 1);
#endif NODUP2
#endif LWSRV_LPR_LOG
    }
#ifdef TIOCNOTTY
    if ((i = open("/dev/tty",2)) > 0) {
      ioctl(i, TIOCNOTTY,(char *) 0);
      close(i);
    }
#endif TIOCNOTTY
#ifdef xenix5
    /*
     * USG process groups:
     * The fork guarantees that the child is not a process group leader.
     * Then setpgrp() can work, whick loses the controllong tty.
     * Note that we must be careful not to be the first to open any tty,
     * or it will become our controlling tty.  C'est la vie.
     */
    setpgrp();
#endif xenix5
  }

/*  dbug.db_pap = TRUE; */
  
  abInit(FALSE);		/* initialize appletalk driver */
  nbpInit();			/* initialize NBP routines */
  PAPInit();			/* init PAP printer routines */

  fprintf(stderr,"lwsrv: Spooler starting for %s.%s.*\n",prtname,prttype);
  if (!spool_setup(tracefile, fontfile, prtname, dictdir)) {
    usage(argv[0]);
    exit(1);
  }
  sprintf(buf,"%s:%s@*",prtname,prttype);

  cpyc2pstr(statbuff.StatusStr,"status: idle");
  err = SLInit(&srefnum, buf, 8, &statbuff);
  if (err != noErr) {
    fprintf(stderr,"lwsrv: SLInit failed: %d\n",err);
    exit(1);
  }

  if (!singlefork)
    signal(SIGCHLD, childdone);

  while (TRUE) {
#ifdef LWSRV_AUFS_SECURITY
    AddrBlock addr;			/* budd */
#endif LWSRV_AUFS_SECURITY
    NewStatus("idle");
    err = GetNextJob(srefnum, &cno, &rcomp);
    if (err != noErr) {
      fprintf(stderr,"lwsrv: GetNextJob failed: %d\n",err);
      exit(1);
    }
    while (rcomp > 0)
      abSleep(20,TRUE);
    fprintf(stderr,"lwsrv: Starting print job for %s\n",prtname);
#ifdef LWSRV_AUFS_SECURITY
    /*** budd.... */
    if ((err = PAPGetNetworkInfo(cno, &addr)) != noErr)
      fprintf(stderr, "Get Network info failed with error %d", err);
    else
      fprintf(stderr,"Connection %d from [Network %d.%d, node %d, socket %d]\n",
	      cno, nkipnetnumber(addr.net),
	      nkipsubnetnumber(addr.net),
	      addr.node, addr.skt);

    if( aufsdb != NULL ) {
	char fname[ 256 ];
	char filename[ 256 ];
	int f, cc, ok;
	struct passwd *pw;
	struct stat statbuf;
	ok = 0;				/* false */
	make_userlogin( filename, aufsdb, addr );
	if( (f = open( filename, 0)) >= 0 &&
	   (cc = read( f, fname, sizeof( fname )-1 )) > 0 ) {
	    if( fname[cc-1] == '\n' )
		fname[cc-1] = '\0';
	    fprintf( stderr, "Found username in aufsdb (%s): %s\n",
		    aufsdb, fname );

	    if ((tempbin = rindex(fname,':')) != NULL) {
	      *tempbin='\0';
	      tempbin++;
	      bin=(char *) malloc(strlen(tempbin)+1);
	      strcpy(bin,tempbin);
	    }
	    
	    (void) stat(filename, &statbuf);
	    if ( (pw = getpwuid((int) statbuf.st_uid)) != NULL ) {
	      if ( strcmp(pw->pw_name, fname) == 0) {
		requid = pw->pw_uid;
		reqgid = pw->pw_gid;
#ifdef LWSRV_LPR_LOG
		requname = pw->pw_name;
#endif LWSRV_LPR_LOG
		ok = 1;
	      }
	    }
	    else
	      if( (pw = getpwnam( fname )) != NULL ) {
		requid = pw->pw_uid;
		reqgid = pw->pw_gid;
#ifdef LWSRV_LPR_LOG
		requname = pw->pw_name;
#endif LWSRV_LPR_LOG
		ok = 1;		/* true */
	      } /* pwnam ok */
	    close(f);  /* close the file */
	  } /* open and read OK */
	if( !ok ) {	/* dump connection with error message */
	    char message[ 512 ];
	    fprintf( stderr, "No username (or invalid) confirmation.\n");
	    gethostname( fname, sizeof( fname )-1 ); /* ick */

	    /* THIS DOES NOT WORK! */
	    sprintf(message, "error: please login to host %s via appleshare",
		    fname );
	    NewStatus( message );
	    sleep( 3 );			/* bleh!! */
/*	    PAPShutdown( cno );		/*  */
	    PAPClose( cno, TRUE );	/* arg2 is ignored!! */
					/* what does it mean? */

	    continue;
	} /* not ok */
    } /* have aufsdb */
    /* ....budd ***/
#endif LWSRV_AUFS_SECURITY

    if (!singlefork) {
      if (fork() != 0) {
	PAPShutdown(cno);	/* kill off connection */
	continue;
      }
      SLClose(srefnum);		/* close server for child */
      nbpShutdown();		/* shutdown nbp for child */
    } else NewStatus("busy, processing job");
    /* need for multi-forking, nice for single forking */
    /* handle the connection in cno */
    signal(SIGCHLD,SIG_IGN);
    childjob(p_opn(cno, BUFMAX)); 
    if (!singlefork)
      exit(0);
  }
}

export
childjob(pf)
PFILE *pf;
{
  long t;
  int argc, i;
  FILE *outfile;
  char tname[100],status[80];
  char pbuf[100],rhbuf[16],jbuf[100];
  char *childargv[20];
#ifdef LWSRV_AUFS_SECURITY
  char bbuf[64];
#endif LWSRV_AUFS_SECURITY
#ifdef NeXT
  char dpistring[6];
#endif NeXT
#ifdef WIFEXITED
  int waitret;
  union wait waitstatus;
#endif WIFEXITED
#ifdef AUTHENTICATE
  register PAPSOCKET *ps;
  register unsigned net;

  if((ps = cnotopapskt(pf->p_cno)) == NULL || !authenticate(net =
   ntohs(ps->addr.net), ps->addr.node)) {
    p_cls(pf);				/* close out the pap connection */
    (void)time(&t);
    fprintf(stderr,"lwsrv: Authentication failed: net %u.%u node %u; on %s",
     ((unsigned)net >> 8), (unsigned)(net & 0xff), (unsigned)ps->addr.node,
     ctime(&t));
    return;
  }
#endif AUTHENTICATE
  
  jobname[0] = username[0] = '\0';

  if (tracefile != NULL)		/* is this a trace? */
    strcpy(tname,tracefile);		/* yes... then output is tracefile */
  else {				/* otherwise use a temp file */
    strcpy(tname,TEMPFILE);
    mktemp(tname);
  }
  
  if ((outfile = fopen(tname, "w+")) == NULL) {
    perror(tname);
  }

  if (singlefork)
    NewStatus("initializing");
  sprintf(status,"receiving job for %s",prtname);

  scandicts(dictdir);

#ifdef PAGECOUNT
  pagecount = -1;
  pcopies = 0;
#endif PAGECOUNT

  while (getjob(pf,outfile)) {	/* while still open... */
    if (singlefork)
      NewStatus(status);
    /* don't send out real eof - causes real problems */
    /* given that we are prepending procsets */
    fprintf(outfile,"%% *EOF*\n");
  }

  fclose(outfile);

  (void) time(&t);

  if (tracefile != NULL)
    fprintf(stderr,"lwsrv: Tracing to file: %s; job %s; user %s; on %s\n",
	    tracefile,jobname,username,ctime(&t));
  else {
    if (rflag)
      fprintf(stderr,"lwsrv: Preserving file in %s\n",tname);
    fprintf(stderr,"lwsrv: Printing job: %s; user %s; on %s\n",
	    jobname,username,ctime(&t));

#ifndef LPRCMD
# ifdef xenix5
#  define LPRCMD "/usr/bin/lpr"
# else xenix5
#  define LPRCMD "/usr/ucb/lpr"
# endif xenix5
#endif  LPRCMD

/*
 * this way lies madness ...
 */

	argc = 0;
	childargv[argc++]="lpr";
#ifdef MELBOURNE
	childargv[argc++]="-v";
#endif MELBOURNE
#ifdef xenix5
	childargv[argc++]="-og";
	sprintf(pbuf,"-d%s",unixpname); /* name of the printer */
	childargv[argc++]=pbuf;
#else xenix5
	sprintf(pbuf,"-P%s",unixpname); /* name of the printer */
	childargv[argc++]=pbuf;
	childargv[argc++]="-J";
#ifdef DOCNAME
	sprintf(jbuf,"MacUser: %s Job: %s",username,jobname); /* job name */
#else  DOCNAME
	sprintf(jbuf,"MacUser: %s",username); /* job name */
#endif DOCNAME
#ifdef PAGECOUNT
	if (pagecount >= 0) {
	  if (pcopies <= 0)
	    pcopies = 1;
	  sprintf(&buf[strlen(buf)], " Pages: %d", pcopies*pagecount);
	}
#endif PAGECOUNT
	childargv[argc++]=jbuf;
#endif xenix5
#ifdef LWSRV_AUFS_SECURITY
	if (aufsdb && bin != NULL) {
#ifdef RUTGERS
	  sprintf(bbuf, "-B%s", bin);
#else  RUTGERS
	  sprintf(bbuf, "-C%s", bin);
#endif RUTGERS
	  childargv[argc++]=bbuf;
	}
#endif LWSRV_AUFS_SECURITY
#ifdef NeXT
        if (nextdpi) {
	  sprintf(dpistring, "-R%s", nextdpi);
	  childargv[argc++]=dpistring;
	}
#endif NeXT
#ifdef xenix5
	/* will this work ... ? */
	sprintf(rhbuf,"-%s%s",rflag ? "" : "r",hflag ? "" : "ob");
#else  xenix5
	sprintf(rhbuf,"-%s%s",rflag ? "" : "r",hflag ? "" : "h");
#endif xenix5
	if (rhbuf[1] != '\0') 
	  childargv[argc++]=rhbuf;	/* include h and/or r flags */
        childargv[argc++]=tname;	/* our temporary file name */
        childargv[argc]=(char *) 0;	/* end of argument list */

	switch (fork()) {
	case 0:
#ifdef LWSRV_AUFS_SECURITY
	    if (aufsdb) {
	/*
	 * dissassociate from any tty to make sure
	 * lpr uses our requid instead of getlogin()
	 *
	 */
#ifndef LWSRV_LPR_LOG
		for(i = 0 ; i < 10 ; i++)
	          (void) close(i);
	 	(void) open("/", 0);
		(void) dup2(0, 1);
		(void) dup2(0, 2);
#else LWSRV_LPR_LOG
		for(i = 3 ; i < 10 ; i++)
		  (void) close(i);
	/* 
	 * stderr is the logfile; stdin and stdout are already 
	 * using /dev/null, therefore disassociated from any tty;
	 * The Rutgers hack requires stdout to be the logfile if
	 * one exists.
	 *
	 */
		/* (void) open("/", 0);	*/
		/* (void) dup2(0, 1);	*/
		/* (void) dup2(0, 2);	*/
#endif LWSRV_LPR_LOG
		chown(tname, requid, reqgid);
		setuid(requid);
	    }
#endif LWSRV_AUFS_SECURITY
#ifdef LWSRV_LPR_LOG
	/*
	 * log all lpr invocations for troubleshooting
	 * printing problems
	 *
	 */
	    fprintf(stderr, "Invoking lpr as user %s using execv: %s",
	        (requname) ? requname : "<unknown>", LPRCMD);
	    for ( i = 1 ; i < argc ; i++) 
	      fprintf(stderr," %s",childargv[i]);
	    fprintf(stderr, "\n");
#endif LWSRV_LPR_LOG
	    if (execv(LPRCMD,childargv)) {
		fprintf(stderr,"exec of %s failed\n", LPRCMD);
		perror("exec");
		exit(-1);
	    }
	    break;
	case -1:
	    fprintf(stderr,"fork failed trying to run %s\n", LPRCMD);
	    perror("fork");
	    break;
	default:
#ifdef WIFEXITED
	/*
	 * This code is important, since lwsrv relies upon lpr to remove
	 * the temporary files, we must remove them if lpr fails for any
	 * reason (unless rflag TRUE). We need the System V equivalent below.
	 *
	 */
	    waitret = wait(&waitstatus); /* wait for the lpr to do its thing */
	    if (waitret == -1 ) {
		perror("wait");
		exit(-1);
	    }
	    if (WIFEXITED(waitstatus)) {
		if ( (int) waitstatus.w_retcode != 0 ) {
		  fprintf(stderr,"lpr exited with status %d, %sremoving %s\n",
			(int) waitstatus.w_retcode, (rflag)?"not ":"",tname);
		    if (!rflag)
		      unlink(tname);
		}
	    }
#else  WIFEXITED
	    /* wait for the lpr to do its thing; using System V wait syntax */
	    { int status, retval, exitval;
  	      retval = wait(&status);
 	      if (retval == -1 ) {
		perror("wait");
		exit(-1);
	      }
	      status &= 0x0000ffff;
	      if ( !(status & 0x000000ff) ) {
		exitval = status & 0x0000ff00;
		if ( exitval != 0 ) {
		   fprintf(stderr,"lpr exited with status %d, %sremoving %s\n",
			exitval, (rflag)?"not ":"",tname);
	 	   if (!rflag)
		      unlink(tname);
		}
	      }
	    }
#endif WIFEXITED
	    break;
	}
  }
  p_cls(pf);				/* close out the pap connection */
}

export void
setjobname(ts)
char *ts;
{
  strcpy(jobname, ts);
}

setusername(ts)
char *ts;
{
  strcpy(username, ts);
}

export
NewStatus(status)
char *status;
{
  char tmp[255];

  if (*username != '\0')
    sprintf(tmp,"job: %s for %s; status: %s",
	    jobname,username,status);
  else
    sprintf(tmp,"status: %s",status);
  cpyc2pstr(statbuff.StatusStr, tmp);
  abSleep(0,TRUE);			/* make sure protocol runs */
}

#ifdef LWSRV_AUFS_SECURITY
/**************** budd... ****************/
/* duplicated in aufs.c and lwsrv.c sigh */
make_userlogin( buf, dir, addrb )
    char *buf, *dir;
    AddrBlock addrb;
{
    sprintf( buf, "%s/net%d.%dnode%d",
	    dir,
	    nkipnetnumber(addrb.net), nkipsubnetnumber(addrb.net),
	    addrb.node
	    );
}
#endif LWSRV_AUFS_SECURITY
